%% Erstellen von Wellenvideos mit optionalem Schnitt
%
% 2011 von Dominik Zobel
%
% Für eine frei wählbare Höhenfunktion wird über Rotation eine kreisförmige Verteilung erzeugt.
% Die Höhenfunktion ist (zumindest: sollte) räumlich und zeitlich beliebig definierbar sein.
% Außerdem ist es möglich, mehrere Überlagerungen zu berechnen und als Höhenfunktion zu übergeben

clear all
close all


%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% --- Grundlegendes --- %%

% Anzahl an Frames pro Sekunde
fps=25;

% Abspieldauer in Sekunden
dauer=8;

% Abspielgeschwindigkeit
geschwindigkeit=1;

% Video in Datei speichern 0: nein, 1: ja
speichern=1;


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% --- Definiere den gewünschten Bereich der auszugebenden Matrix --- %%

% Die Dimension der Wellenmatrix festlegen. Je größer, desto länger die Berechnung
anzahl_an_zeilen=160;
anzahl_an_spalten=160;

% Einen Bereich festlegen, der NICHT geplottet werden soll. Dazu sind zwei verschiedene Winkelangaben nötig.
% Bei zwei gleichen Winkeln wird alles geplottet
phi_start=270;
phi_end=360;

% Schnittkante über Einzelpunkte zeichnen?
einzel_schnittkante=1;

% Größtmöglichen Kreis ausschneiden (1) oder die Matrix als Rechteck ausgeben (0)?
kreis_im_graph=0;


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% --- Definition der Höhenfunktion --- %%

% Funktionswerte für Anzahl an Werten (alternativ auch nur eine Periode bei gleichbleibenden Funktionen)
% und ihr zeitlicher Verlauf in den Spalten. Dazu wird ein Schritte pro Feldabstand definiert
% (Glattheit der Funktion - je größer desto glatter; moderate Beeinflussung der Rechenzeit)
schrittweitenaufloesung=10;

% Erstelle einen Vektor, der in Zehntelschritten bis zum Eckwert der Matrix geht.
% Der Eckwert wird immer als nächsthöherer Integerwert angegeben.
x=0:1/schrittweitenaufloesung:ceil(sqrt(anzahl_an_zeilen^2+anzahl_an_spalten^2))-1;

daempfungsfaktor_zeitlich=4;
daempfungsfaktor_ausbreitung=20;
periodenlaenge=50;

disp(['Berechne Höhen- und Wellenmatrix ...']);

Hoehenfunktion=[];
for zeitschritt=0:1:dauer*fps-1

   % Eine wachsende Sprungfunktion
   %funktion=ones(size(x));
   %start_point=round(ablaufgeschwindigkeit*zeitschritt/fps*schrittweitenaufloesung);
   %if (start_point > max(size(x))-1)
   %   start_point=max(size(x))-1;
   %end
   %funktion(1+start_point:max(size(x)))=0;


   % Ein impulsähnliches Signal
   %funktion=zeros(size(x));
   %start_point=round(ablaufgeschwindigkeit*zeitschritt/fps*schrittweitenaufloesung);
   %end_point=start_point+ceil(20*exp(zeitschritt*daempfungsfaktor_zeitlich/(dauer*fps)));
   %if (start_point > max(size(x)))
   %   start_point=max(size(x));
   %end
   %if (end_point > max(size(x)))
   %   end_point=max(size(x));
   %end
   %funktion(1+start_point:end_point)=exp(-zeitschritt*daempfungsfaktor_zeitlich/(dauer*fps));


   % Ein abklingender Sinus mit Entstehung
   funktion=exp(-zeitschritt*daempfungsfaktor_zeitlich/(dauer*fps))*exp(-x*daempfungsfaktor_ausbreitung*2/max(size(x))).*sin(pi*4/periodenlaenge*(x-zeitschritt*geschwindigkeit));
   start_point=round(zeitschritt*10*geschwindigkeit);
   if (start_point < max(size(funktion)))
      funktion(1+start_point:max(size(funktion)))=0;
   end
   clear start_point

   Hoehenfunktion=[Hoehenfunktion;funktion];
end

clear daempfungsfaktor_zeitlich daempfungsfaktor_ausbreitung ausbreitungsgeschwindigkeit
%clear x zeitschritt funktion


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% --- Berechnung der Schablonen zum Ausschneiden --- %%

% Um das Zentrum bei geraden und ungeraden Zeilen/Spaltenzahlen optimal zu positionieren wird folgendes verwendet:
zentrum_abstand=[(anzahl_an_zeilen)/2 (anzahl_an_spalten)/2];

% Matlab fängt ab 1 zu zählen an. Wenn also direkt auf das Zentrum der Matrix zugegriffen werden soll, dann benutze folgendes:
zentrum_matrixkoordinaten=[(anzahl_an_zeilen+1)/2 (anzahl_an_spalten+1)/2];


radius=min(zentrum_abstand(1),zentrum_abstand(2));

Kreisschablone=zeros(anzahl_an_zeilen,anzahl_an_spalten);
Schnittschablone=zeros(anzahl_an_zeilen,anzahl_an_spalten);
Randschablone=Inf(anzahl_an_zeilen,anzahl_an_spalten);
Winkelmatrix=zeros(anzahl_an_zeilen,anzahl_an_spalten);

if (phi_start ~= phi_end)
   for zeile=1:anzahl_an_zeilen
      for spalte=1:anzahl_an_spalten
         % Bestimme den Winkel zwischen dem Mittleren Punkt und den Feldpunkten in Grad
         winkel=180/pi*atan((zentrum_matrixkoordinaten(1)-zeile)/(spalte-zentrum_matrixkoordinaten(2)));
         if ((spalte-zentrum_matrixkoordinaten(2)) >= 0)
            winkel=winkel+180;
         end
         if (((spalte-zentrum_matrixkoordinaten(2)) < 0) & ((zeile-zentrum_matrixkoordinaten(1)) < 0))
            winkel=winkel+360;
         end
         % Zur Überprüfung kann hier eine Winkelmatrix erstellt und betrachtet werden
         Winkelmatrix(zeile,spalte)=winkel;

         if (((winkel > phi_start) & (winkel < phi_end)) | ((zeile == zentrum_matrixkoordinaten(1)) && (spalte == zentrum_matrixkoordinaten(2))))
            Schnittschablone(zeile,spalte)=Inf;

            Randschablone(zeile,spalte)=0;

            % Die ersten vier sind für ein + Element
            if (zeile+1 <= anzahl_an_zeilen)
               Randschablone(zeile+1,spalte)=0;
            end
            if (zeile-1 >= 1)
               Randschablone(zeile-1,spalte)=0;
            end
            if (spalte+1 <= anzahl_an_spalten)
               Randschablone(zeile,spalte+1)=0;
            end
            if (spalte-1 >= 1)
               Randschablone(zeile,spalte-1)=0;
            end

            % Die nächsten vier ergänzen das + zum Vollblockelement
            if ((zeile+1 <= anzahl_an_zeilen) & (spalte+1 <= anzahl_an_spalten))
               Randschablone(zeile+1,spalte+1)=0;
            end
            if ((zeile+1 <= anzahl_an_zeilen) & (spalte-1 >= 1))
               Randschablone(zeile+1,spalte-1)=0;
            end
            if ((zeile-1 >= 1) & (spalte+1 <= anzahl_an_spalten))
               Randschablone(zeile-1,spalte+1)=0;
            end
            if ((zeile-1 >= 1) & (spalte-1 >= 1))
               Randschablone(zeile-1,spalte-1)=0;
            end
         end
      end
   end
end

if (kreis_im_graph == 1)
   for zeile=1:anzahl_an_zeilen
      for spalte=1:anzahl_an_spalten
         if (radius < round(sqrt((zeile-zentrum_matrixkoordinaten(1))^2+(spalte-zentrum_matrixkoordinaten(2))^2)))
            Kreisschablone(zeile,spalte)=Inf;
         end
      end
   end
end

Schnittschablone=Schnittschablone+Kreisschablone;
Randschablone=Randschablone+Schnittschablone;
clear winkel Kreisschablone


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% --- Berechnen der Wellenmatrix aus der Höhenfunktion --- %%


Wellenmatrix=zeros(anzahl_an_zeilen,anzahl_an_spalten,dauer*fps);

for zeitschritt=1:dauer*fps
   for zeile=1:anzahl_an_zeilen
      for spalte=1:anzahl_an_spalten
         % Für jeden Eintrag der Wellenmatrix wird berechnet, wie weit der Punkt vom Mittelpunkt entfernt ist.
         % Anhand der Entfernung wird ein Wert aus der feiner aufgelösten Hoehenfunktion zugewiesen
         Wellenmatrix(zeile,spalte,zeitschritt)=Hoehenfunktion(zeitschritt,1+round(schrittweitenaufloesung*sqrt((zeile-zentrum_matrixkoordinaten(1))^2+(spalte-zentrum_matrixkoordinaten(2))^2)));
      end
   end

%   for zeile=1:anzahl_an_zeilen
%      for spalte=1:anzahl_an_spalten
%         % Für jeden Eintrag der Wellenmatrix wird berechnet, wie weit ein weiterer Punkt vom Mittelpunkt entfernt ist.
%         % Anhand der Entfernung wird ein Wert aus der feiner aufgelösten Hoehenfunktion zugewiesen
%         Wellenmatrix(zeile,spalte,zeitschritt)=Wellenmatrix(zeile,spalte,zeitschritt)+Hoehenfunktion(zeitschritt,1+round(schrittweitenaufloesung*sqrt((zeile+20-zentrum_matrixkoordinaten(1))^2+(spalte-30-zentrum_matrixkoordinaten(2))^2)));
%      end
%   end

   % Schneide Fenster aus den Daten heraus, falls gefordert
   Wellenmatrix(:,:,zeitschritt)=squeeze(Wellenmatrix(:,:,zeitschritt))+Schnittschablone;
end

clear zeile spalte zeitschritt

schnittfunktionsaufloesung=schrittweitenaufloesung^2;

% Sind die Winkel unterschiedlich? Dann werden die Schnittfunktionen gezeichnet
if (phi_start ~= phi_end)

   x1=[];y1=[];x2=[];y2=[];
   laenge1=[];laenge2=[];

   % Der eine Fall beschreibt, wenn nur ein Kreisausschnitt betrachtet wird. In diesem Fall sind die Beziehungen zueinander einfach
   if (kreis_im_graph == 1)
      if (phi_start < 0) || (phi_start > 360) || (phi_end < 0) || (phi_end > 360)
         disp('Winkel zwischen 0 und 360 Grad wählen');
         return;
      end
      for zaehlbit=1:schnittfunktionsaufloesung
         x1(zaehlbit)=zentrum_matrixkoordinaten(2)-cos(phi_start*pi/180)*zaehlbit/schnittfunktionsaufloesung*radius;
         y1(zaehlbit)=zentrum_matrixkoordinaten(1)+sin(phi_start*pi/180)*zaehlbit/schnittfunktionsaufloesung*radius;
         x2(zaehlbit)=zentrum_matrixkoordinaten(2)-cos(phi_end*pi/180)*zaehlbit/schnittfunktionsaufloesung*radius;
         y2(zaehlbit)=zentrum_matrixkoordinaten(1)+sin(phi_end*pi/180)*zaehlbit/schnittfunktionsaufloesung*radius;
      end
      laenge1=radius;
      laenge2=radius;
   else

      % In diesem Zweig der Schleife wird kein Kreisausschnitt, sondern ein beliebiges Rechteck betrachtet.
      % Dadurch ergeben sich unterschiedliche Herausforderungen. Zum einen ist es wichtig zu wissen, auf welcher
      % Seite die Schnittfunktion den Rand schneidet, da sin und cos Argumente nur jeweils für Werte einer Kante (der Matrix)
      % angewendet werden können. Zuerst wird der Startwinkel betrachtet und die dazugehörigen Koordinaten bereitgestellt
      phi_max=180/pi*atan(anzahl_an_zeilen/anzahl_an_spalten);

      if (phi_start < phi_max)
         for zaehlbit=1:schnittfunktionsaufloesung
            x1(zaehlbit)=zentrum_matrixkoordinaten(2)-(zaehlbit-1)/schnittfunktionsaufloesung*(zentrum_abstand(2)-0.5);
            y1(zaehlbit)=zentrum_matrixkoordinaten(1)+tan(phi_start*pi/180)*(zaehlbit-1)/schnittfunktionsaufloesung*(zentrum_abstand(2)-0.5);
         end
         laenge1=(zentrum_abstand(2)-0.5)/cos(phi_start*pi/180);
      else
         if (phi_start < 180-phi_max)
            for zaehlbit=1:schnittfunktionsaufloesung
               x1(zaehlbit)=zentrum_matrixkoordinaten(2)+tan((phi_start-90)*pi/180)*(zaehlbit-1)/schnittfunktionsaufloesung*(zentrum_abstand(1)-0.5);
               y1(zaehlbit)=zentrum_matrixkoordinaten(1)+(zaehlbit-1)/schnittfunktionsaufloesung*(zentrum_abstand(1)-0.5);
            end
            laenge1=(zentrum_abstand(1)-0.5)/cos((phi_start-90)*pi/180);
         else
            if (phi_start < 180+phi_max)
               for zaehlbit=1:schnittfunktionsaufloesung
                  x1(zaehlbit)=zentrum_matrixkoordinaten(2)+(zaehlbit-1)/schnittfunktionsaufloesung*(zentrum_abstand(2)-0.5);
                  y1(zaehlbit)=zentrum_matrixkoordinaten(1)-tan((phi_start-180)*pi/180)*(zaehlbit-1)/schnittfunktionsaufloesung*(zentrum_abstand(2)-0.5);
               end
               laenge1=(zentrum_abstand(2)-0.5)/cos((phi_start-180)*pi/180);
            else
               if (phi_start < 360-phi_max)
                  for zaehlbit=1:schnittfunktionsaufloesung
                     x1(zaehlbit)=zentrum_matrixkoordinaten(2)-tan((phi_start-270)*pi/180)*(zaehlbit-1)/schnittfunktionsaufloesung*(zentrum_abstand(1)-0.5);
                     y1(zaehlbit)=zentrum_matrixkoordinaten(1)-(zaehlbit-1)/schnittfunktionsaufloesung*(zentrum_abstand(1)-0.5);
                  end
                  laenge1=(zentrum_abstand(1)-0.5)/cos((phi_start-270)*pi/180);
               else
                  if (phi_start <= 360)
                     for zaehlbit=1:schnittfunktionsaufloesung
                        x1(zaehlbit)=zentrum_matrixkoordinaten(2)-(zaehlbit-1)/schnittfunktionsaufloesung*(zentrum_abstand(2)-0.5);
                        y1(zaehlbit)=zentrum_matrixkoordinaten(1)+tan((phi_start-360)*pi/180)*(zaehlbit-1)/schnittfunktionsaufloesung*(zentrum_abstand(2)-0.5);
                     end
                     laenge1=(zentrum_abstand(2)-0.5)/cos(phi_start*pi/180);

                   else
                      disp('Winkel zwischen 0 und 360 Grad wählen');
                      return;
                   end
               end
            end
         end
      end

      if (phi_end < phi_max)
         for zaehlbit=1:schnittfunktionsaufloesung
            x2(zaehlbit)=zentrum_matrixkoordinaten(2)-(zaehlbit-1)/schnittfunktionsaufloesung*(zentrum_abstand(2)-0.5);
            y2(zaehlbit)=zentrum_matrixkoordinaten(1)+tan(phi_end*pi/180)*(zaehlbit-1)/schnittfunktionsaufloesung*(zentrum_abstand(2)-0.5);
         end
         laenge2=(zentrum_abstand(2)-0.5)/cos(phi_end*pi/180);
      else
         if (phi_end < 180-phi_max)
            for zaehlbit=1:schnittfunktionsaufloesung
               x2(zaehlbit)=zentrum_matrixkoordinaten(2)+tan((phi_end-90)*pi/180)*(zaehlbit-1)/schnittfunktionsaufloesung*(zentrum_abstand(1)-0.5);
               y2(zaehlbit)=zentrum_matrixkoordinaten(1)+(zaehlbit-1)/schnittfunktionsaufloesung*(zentrum_abstand(1)-0.5);
            end
            laenge2=(zentrum_abstand(1)-0.5)/cos((phi_end-90)*pi/180);
         else
            if (phi_end < 180+phi_max)
               for zaehlbit=1:schnittfunktionsaufloesung
                  x2(zaehlbit)=zentrum_matrixkoordinaten(2)+(zaehlbit-1)/schnittfunktionsaufloesung*(zentrum_abstand(2)-0.5);
                  y2(zaehlbit)=zentrum_matrixkoordinaten(1)-tan((phi_end-180)*pi/180)*(zaehlbit-1)/schnittfunktionsaufloesung*(zentrum_abstand(2)-0.5);
               end
               laenge2=(zentrum_abstand(2)-0.5)/cos((phi_end-180)*pi/180);
            else
               if (phi_end < 360-phi_max)
                  for zaehlbit=1:schnittfunktionsaufloesung
                     x2(zaehlbit)=zentrum_matrixkoordinaten(2)-tan((phi_end-270)*pi/180)*(zaehlbit-1)/schnittfunktionsaufloesung*(zentrum_abstand(1)-0.5);
                     y2(zaehlbit)=zentrum_matrixkoordinaten(1)-(zaehlbit-1)/schnittfunktionsaufloesung*(zentrum_abstand(1)-0.5);
                  end
                  laenge2=(zentrum_abstand(1)-0.5)/cos((phi_end-270)*pi/180);
               else
                  if (phi_end <= 360)
                     for zaehlbit=1:schnittfunktionsaufloesung
                        x2(zaehlbit)=zentrum_matrixkoordinaten(2)-(zaehlbit-1)/schnittfunktionsaufloesung*(zentrum_abstand(2)-0.5);
                        y2(zaehlbit)=zentrum_matrixkoordinaten(1)+tan((phi_end-360)*pi/180)*(zaehlbit-1)/schnittfunktionsaufloesung*(zentrum_abstand(2)-0.5);
                     end
                     laenge2=(zentrum_abstand(2)-0.5)/cos(phi_end*pi/180);
                  else
                     disp('Winkel zwischen 0 und 360 Grad wählen');
                     return;
                  end
               end
            end
         end
      end
      clear phi_max
   end
end




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% --- Grafische Ausgabe --- %%

disp(['Starte grafische Ausgabe...']);

%% Grafische Ausgabe
set(0,'DefaultTextInterpreter','latex')

% Ein Fenster öffnen und zuweisen
fig_handle=figure(1);

img_handle=axes;

color=[zeros(1,16) 0:1/31:1 ones(1,16)];

% Blautöne
cmap=[color'/2.2 color'/1.2 color'];

% Grüntöne
%cmap=[color'*0.2 color'*0.7 color'*0.4];

set(fig_handle,'Position',[0 0 540 360],'Color',[1 1 1],'colormap',cmap)
set(img_handle,'CLim',[-1 1],'FontName','Helvetica')


for zeitlaeufer=1:size(Wellenmatrix,3)
   mesh(img_handle,squeeze(Wellenmatrix(:,:,zeitlaeufer)))
   set(fig_handle,'colormap',cmap)

   % Mit dezenten Werten
   %set(img_handle,'CLim',[-1 1],'ZLim',[-1 1],'XColor',[0.3 0.3 0.3],'YColor',[0.3 0.3 0.3],'ZColor',[0.3 0.3 0.3],'FontName','Helvetica')

   % Ohne irgendetwas
   set(img_handle,'CLim',[-1 1],'ZLim',[-1 1],'XColor',[1 1 1],'YColor',[1 1 1],'ZColor',[1 1 1],'CameraPosition',[-anzahl_an_zeilen*2.5 -anzahl_an_spalten*2 3],'CameraViewAngle',12)

   %set(img_handle,'CameraPosition',[60 170 1])

   if (phi_start ~= phi_end)
      hold(img_handle,'on')
      if (einzel_schnittkante == 1)
         for horiz_randlaeufer=1:size(Randschablone,1)
            for vert_randlaeufer=1:size(Randschablone,2)
               if (~isinf(Randschablone(horiz_randlaeufer,vert_randlaeufer)))
                  plot3(img_handle,vert_randlaeufer,horiz_randlaeufer,Wellenmatrix(horiz_randlaeufer,vert_randlaeufer,zeitlaeufer),'Color',[1 0 0],'Marker','.','Markersize',8)
               end
            end
         end
      else
         z1=Hoehenfunktion(zeitlaeufer,ceil([1:laenge1*schrittweitenaufloesung/(schnittfunktionsaufloesung-1):laenge1*schrittweitenaufloesung+1]));
         z2=Hoehenfunktion(zeitlaeufer,ceil([1:laenge2*schrittweitenaufloesung/(schnittfunktionsaufloesung-1):laenge2*schrittweitenaufloesung+1]));
         plot3(x1,y1,z1,'Color',[0.3 0.8 0.5],'Linewidth',2)
         plot3(x2,y2,z2,'Color',[0.3 0.8 0.5],'Linewidth',2)
      end
      hold(img_handle,'off')
   end


   Filmmatrix(zeitlaeufer)=getframe(fig_handle);
end

%% Erstelle die Videodatei
if (speichern == 1)
   disp(['Schreibe Video...']);
   movie2avi(Filmmatrix,'test.avi','fps',fps)
end

%% [Extern] Komprimiere die Videodatei mit ffmpeg
% ffmpeg -sameq -i test.avi neu.avi

disp(['Beendet']);